# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Helpers for intersecting B |eacute| zier curves via geometric methods.

The functions here are pure Python and many have equivalent implementations
written in Fortran and exposed via a Cython wrapper.

.. |eacute| unicode:: U+000E9 .. LATIN SMALL LETTER E WITH ACUTE
   :trim:
"""

import itertools

import numpy as np

from bezier import _py_curve_helpers
from bezier import _py_helpers
from bezier import _py_intersection_helpers


# Set the threshold for exponent at half the bits available, this way one round
# of Newton's method can (usually) finish the job by squaring the error.
_ERROR_VAL = 0.5 ** 26
_MAX_INTERSECT_SUBDIVISIONS = 20
_MAX_CANDIDATES = 64
_UNHANDLED_LINES = (
    "If both curves are lines, the intersection should have "
    "been computed already."
)
_TOO_MANY_TEMPLATE = (
    "The number of candidate intersections is too high.\n"
    "{:d} candidate pairs."
)
_NO_CONVERGE_TEMPLATE = (
    "Curve intersection failed to converge to approximately linear "
    "subdivisions after {:d} iterations."
)
_MIN_INTERVAL_WIDTH = 0.5 ** 40


def bbox_intersect(nodes1, nodes2):
    r"""Bounding box intersection predicate.

    .. note::

       There is also a Fortran implementation of this function, which
       will be used if it can be built.

    Determines if the bounding box of two sets of control points
    intersects in :math:`\mathbf{R}^2` with non-trivial
    intersection (i.e. tangent bounding boxes are insufficient).

    .. note::

       Though we assume (and the code relies on this fact) that
       the nodes are two-dimensional, we don't check it.

    Args:
        nodes1 (numpy.ndarray): Set of control points for a
            B |eacute| zier shape.
        nodes2 (numpy.ndarray): Set of control points for a
            B |eacute| zier shape.

    Returns:
        int: Enum from ``BoxIntersectionType`` indicating the type of
        bounding box intersection.
    """
    left1, right1, bottom1, top1 = _py_helpers.bbox(nodes1)
    left2, right2, bottom2, top2 = _py_helpers.bbox(nodes2)
    if right2 < left1 or right1 < left2 or top2 < bottom1 or top1 < bottom2:
        return BoxIntersectionType.DISJOINT

    if (
        right2 == left1
        or right1 == left2
        or top2 == bottom1
        or top1 == bottom2
    ):
        return BoxIntersectionType.TANGENT

    else:
        return BoxIntersectionType.INTERSECTION


def linearization_error(nodes):
    r"""Compute the maximum error of a linear approximation.

    .. note::

       There is also a Fortran implementation of this function, which
       will be used if it can be built.

    .. note::

       This is a helper for :class:`.Linearization`, which is used during the
       curve-curve intersection process.

    We use the line

    .. math::

       L(s) = v_0 (1 - s) + v_n s

    and compute a bound on the maximum error

    .. math::

       \max_{s \in \left[0, 1\right]} \|B(s) - L(s)\|_2.

    Rather than computing the actual maximum (a tight bound), we
    use an upper bound via the remainder from Lagrange interpolation
    in each component. This leaves us with :math:`\frac{s(s - 1)}{2!}`
    times the second derivative in each component.

    The second derivative curve is degree :math:`d = n - 2` and
    is given by

    .. math::

       B''(s) = n(n - 1) \sum_{j = 0}^{d} \binom{d}{j} s^j
       (1 - s)^{d - j} \cdot \Delta^2 v_j

    Due to this form (and the convex combination property of
    B |eacute| zier Curves) we know each component of the second derivative
    will be bounded by the maximum of that component among the
    :math:`\Delta^2 v_j`.

    For example, the curve

    .. math::

       B(s) = \left[\begin{array}{c} 0 \\ 0 \end{array}\right] (1 - s)^2
           + \left[\begin{array}{c} 3 \\ 1 \end{array}\right] 2s(1 - s)
           + \left[\begin{array}{c} 9 \\ -2 \end{array}\right] s^2

    has
    :math:`B''(s) \equiv \left[\begin{array}{c} 6 \\ -8 \end{array}\right]`
    which has norm :math:`10` everywhere, hence the maximum error is

    .. math::

       \left.\frac{s(1 - s)}{2!} \cdot 10\right|_{s = \frac{1}{2}}
          = \frac{5}{4}.

    .. image:: ../images/linearization_error.png
       :align: center

    .. testsetup:: linearization-error, linearization-error-fail

       import numpy as np
       import bezier
       from bezier._py_geometric_intersection import linearization_error

    .. doctest:: linearization-error

       >>> nodes = np.asfortranarray([
       ...     [0.0, 3.0, 9.0],
       ...     [0.0, 1.0, -2.0],
       ... ])
       >>> linearization_error(nodes)
       1.25

    .. testcleanup:: linearization-error

       import make_images
       make_images.linearization_error(nodes)

    As a **non-example**, consider a "pathological" set of control points:

    .. math::

       B(s) = \left[\begin{array}{c} 0 \\ 0 \end{array}\right] (1 - s)^3
           + \left[\begin{array}{c} 5 \\ 12 \end{array}\right] 3s(1 - s)^2
           + \left[\begin{array}{c} 10 \\ 24 \end{array}\right] 3s^2(1 - s)
           + \left[\begin{array}{c} 30 \\ 72 \end{array}\right] s^3

    By construction, this lies on the line :math:`y = \frac{12x}{5}`, but
    the parametrization is cubic:
    :math:`12 \cdot x(s) = 5 \cdot y(s) = 180s(s^2 + 1)`. Hence, the fact
    that the curve is a line is not accounted for and we take the worse
    case among the nodes in:

    .. math::

       B''(s) = 3 \cdot 2 \cdot \left(
           \left[\begin{array}{c} 0 \\ 0 \end{array}\right] (1 - s)
           + \left[\begin{array}{c} 15 \\ 36 \end{array}\right] s\right)

    which gives a nonzero maximum error:

    .. doctest:: linearization-error-fail

       >>> nodes = np.asfortranarray([
       ...     [0.0, 5.0, 10.0, 30.0],
       ...     [0.0, 12.0, 24.0, 72.0],
       ... ])
       >>> linearization_error(nodes)
       29.25

    Though it may seem that ``0`` is a more appropriate answer, consider
    the **goal** of this function. We seek to linearize curves and then
    intersect the linear approximations. Then the :math:`s`-values from
    the line-line intersection is lifted back to the curves. Thus
    the error :math:`\|B(s) - L(s)\|_2` is more relevant than the
    underyling algebraic curve containing :math:`B(s)`.

    .. note::

       It may be more appropriate to use a **relative** linearization error
       rather than the **absolute** error provided here. It's unclear if
       the domain :math:`\left[0, 1\right]` means the error is **already**
       adequately scaled or if the error should be scaled by the arc
       length of the curve or the (easier-to-compute) length of the line.

    Args:
        nodes (numpy.ndarray): Nodes of a curve.

    Returns:
        float: The maximum error between the curve and the
        linear approximation.
    """
    _, num_nodes = nodes.shape
    degree = num_nodes - 1
    if degree == 1:
        return 0.0

    second_deriv = nodes[:, :-2] - 2.0 * nodes[:, 1:-1] + nodes[:, 2:]
    worst_case = np.max(np.abs(second_deriv), axis=1)
    # max_{0 <= s <= 1} s(1 - s)/2 = 1/8 = 0.125
    multiplier = 0.125 * degree * (degree - 1)
    # NOTE: worst_case is 1D due to np.max(), so this is the vector norm.
    return multiplier * np.linalg.norm(worst_case, ord=2)


def segment_intersection(start0, end0, start1, end1):
    r"""Determine the intersection of two line segments.

    Assumes each line is parametric

    .. math::

       \begin{alignat*}{2}
        L_0(s) &= S_0 (1 - s) + E_0 s &&= S_0 + s \Delta_0 \\
        L_1(t) &= S_1 (1 - t) + E_1 t &&= S_1 + t \Delta_1.
       \end{alignat*}

    To solve :math:`S_0 + s \Delta_0 = S_1 + t \Delta_1`, we use the
    cross product:

    .. math::

       \left(S_0 + s \Delta_0\right) \times \Delta_1 =
           \left(S_1 + t \Delta_1\right) \times \Delta_1 \Longrightarrow
       s \left(\Delta_0 \times \Delta_1\right) =
           \left(S_1 - S_0\right) \times \Delta_1.

    Similarly

    .. math::

       \Delta_0 \times \left(S_0 + s \Delta_0\right) =
           \Delta_0 \times \left(S_1 + t \Delta_1\right) \Longrightarrow
       \left(S_1 - S_0\right) \times \Delta_0 =
           \Delta_0 \times \left(S_0 - S_1\right) =
           t \left(\Delta_0 \times \Delta_1\right).

    .. note::

       Since our points are in :math:`\mathbf{R}^2`, the "traditional"
       cross product in :math:`\mathbf{R}^3` will always point in the
       :math:`z` direction, so in the above we mean the :math:`z`
       component of the cross product, rather than the entire vector.

    For example, the diagonal lines

    .. math::

       \begin{align*}
        L_0(s) &= \left[\begin{array}{c} 0 \\ 0 \end{array}\right] (1 - s) +
                  \left[\begin{array}{c} 2 \\ 2 \end{array}\right] s \\
        L_1(t) &= \left[\begin{array}{c} -1 \\ 2 \end{array}\right] (1 - t) +
                  \left[\begin{array}{c} 1 \\ 0 \end{array}\right] t
       \end{align*}

    intersect at :math:`L_0\left(\frac{1}{4}\right) =
    L_1\left(\frac{3}{4}\right) =
    \frac{1}{2} \left[\begin{array}{c} 1 \\ 1 \end{array}\right]`.

    .. image:: ../images/segment_intersection1.png
       :align: center

    .. testsetup:: segment-intersection1, segment-intersection2

       import numpy as np
       from bezier._py_geometric_intersection import segment_intersection

    .. doctest:: segment-intersection1
       :options: +NORMALIZE_WHITESPACE

       >>> start0 = np.asfortranarray([0.0, 0.0])
       >>> end0 = np.asfortranarray([2.0, 2.0])
       >>> start1 = np.asfortranarray([-1.0, 2.0])
       >>> end1 = np.asfortranarray([1.0, 0.0])
       >>> s, t, _ = segment_intersection(start0, end0, start1, end1)
       >>> s
       0.25
       >>> t
       0.75

    .. testcleanup:: segment-intersection1

       import make_images
       make_images.segment_intersection1(start0, end0, start1, end1, s)

    Taking the parallel (but different) lines

    .. math::

       \begin{align*}
        L_0(s) &= \left[\begin{array}{c} 1 \\ 0 \end{array}\right] (1 - s) +
                  \left[\begin{array}{c} 0 \\ 1 \end{array}\right] s \\
        L_1(t) &= \left[\begin{array}{c} -1 \\ 3 \end{array}\right] (1 - t) +
                  \left[\begin{array}{c} 3 \\ -1 \end{array}\right] t
       \end{align*}

    we should be able to determine that the lines don't intersect, but
    this function is not meant for that check:

    .. image:: ../images/segment_intersection2.png
       :align: center

    .. doctest:: segment-intersection2
       :options: +NORMALIZE_WHITESPACE

       >>> start0 = np.asfortranarray([1.0, 0.0])
       >>> end0 = np.asfortranarray([0.0, 1.0])
       >>> start1 = np.asfortranarray([-1.0, 3.0])
       >>> end1 = np.asfortranarray([3.0, -1.0])
       >>> _, _, success = segment_intersection(start0, end0, start1, end1)
       >>> success
       False

    .. testcleanup:: segment-intersection2

       import make_images
       make_images.segment_intersection2(start0, end0, start1, end1)

    Instead, we use :func:`parallel_lines_parameters`:

    .. testsetup:: segment-intersection2-continued

       import numpy as np
       from bezier._py_geometric_intersection import parallel_lines_parameters

       start0 = np.asfortranarray([1.0, 0.0])
       end0 = np.asfortranarray([0.0, 1.0])
       start1 = np.asfortranarray([-1.0, 3.0])
       end1 = np.asfortranarray([3.0, -1.0])

    .. doctest:: segment-intersection2-continued

       >>> disjoint, _ = parallel_lines_parameters(start0, end0, start1, end1)
       >>> disjoint
       True

    .. note::

       There is also a Fortran implementation of this function, which
       will be used if it can be built.

    Args:
        start0 (numpy.ndarray): A 1D NumPy ``2``-array that is the start
            vector :math:`S_0` of the parametric line :math:`L_0(s)`.
        end0 (numpy.ndarray): A 1D NumPy ``2``-array that is the end
            vector :math:`E_0` of the parametric line :math:`L_0(s)`.
        start1 (numpy.ndarray): A 1D NumPy ``2``-array that is the start
            vector :math:`S_1` of the parametric line :math:`L_1(s)`.
        end1 (numpy.ndarray): A 1D NumPy ``2``-array that is the end
            vector :math:`E_1` of the parametric line :math:`L_1(s)`.

    Returns:
        Tuple[float, float, bool]: Pair of :math:`s_{\ast}` and
        :math:`t_{\ast}` such that the lines intersect:
        :math:`L_0\left(s_{\ast}\right) = L_1\left(t_{\ast}\right)` and then
        a boolean indicating if an intersection was found (i.e. if the lines
        aren't parallel).
    """
    delta0 = end0 - start0
    delta1 = end1 - start1
    cross_d0_d1 = _py_helpers.cross_product(delta0, delta1)
    if cross_d0_d1 == 0.0:
        return None, None, False

    else:
        start_delta = start1 - start0
        s = _py_helpers.cross_product(start_delta, delta1) / cross_d0_d1
        t = _py_helpers.cross_product(start_delta, delta0) / cross_d0_d1
        return s, t, True


def parallel_lines_parameters(start0, end0, start1, end1):
    r"""Checks if two parallel lines ever meet.

    Meant as a back-up when :func:`segment_intersection` fails.

    .. note::

       This function assumes but never verifies that the lines
       are parallel.

    In the case that the segments are parallel and lie on **different**
    lines, then there is a **guarantee** of no intersection. However, if
    they are on the exact same line, they may define a shared segment
    coincident to both lines.

    In :func:`segment_intersection`, we utilized the normal form of the
    lines (via the cross product):

    .. math::

       \begin{align*}
       L_0(s) \times \Delta_0 &\equiv S_0 \times \Delta_0 \\
       L_1(t) \times \Delta_1 &\equiv S_1 \times \Delta_1
       \end{align*}

    So, we can detect if :math:`S_1` is on the first line by
    checking if

    .. math::

       S_0 \times \Delta_0 \stackrel{?}{=} S_1 \times \Delta_0.

    If it is not on the first line, then we are done, the
    segments don't meet:

    .. image:: ../images/parallel_lines_parameters1.png
       :align: center

    .. testsetup:: parallel-different1, parallel-different2

       import numpy as np
       from bezier._py_geometric_intersection import parallel_lines_parameters

    .. doctest:: parallel-different1

       >>> # Line: y = 1
       >>> start0 = np.asfortranarray([0.0, 1.0])
       >>> end0 = np.asfortranarray([1.0, 1.0])
       >>> # Vertical shift up: y = 2
       >>> start1 = np.asfortranarray([-1.0, 2.0])
       >>> end1 = np.asfortranarray([3.0, 2.0])
       >>> disjoint, _ = parallel_lines_parameters(start0, end0, start1, end1)
       >>> disjoint
       True

    .. testcleanup:: parallel-different1

       import make_images
       make_images.helper_parallel_lines(
           start0, end0, start1, end1, "parallel_lines_parameters1.png")

    If :math:`S_1` **is** on the first line, we want to check that
    :math:`S_1` and :math:`E_1` define parameters outside of
    :math:`\left[0, 1\right]`. To compute these parameters:

    .. math::

       L_1(t) = S_0 + s_{\ast} \Delta_0 \Longrightarrow
           s_{\ast} = \frac{\Delta_0^T \left(
               L_1(t) - S_0\right)}{\Delta_0^T \Delta_0}.

    For example, the intervals :math:`\left[0, 1\right]` and
    :math:`\left[\frac{3}{2}, 2\right]` (via
    :math:`S_1 = S_0 + \frac{3}{2} \Delta_0` and
    :math:`E_1 = S_0 + 2 \Delta_0`) correspond to segments that
    don't meet:

    .. image:: ../images/parallel_lines_parameters2.png
       :align: center

    .. doctest:: parallel-different2

       >>> start0 = np.asfortranarray([1.0, 0.0])
       >>> delta0 = np.asfortranarray([2.0, -1.0])
       >>> end0 = start0 + 1.0 * delta0
       >>> start1 = start0 + 1.5 * delta0
       >>> end1 = start0 + 2.0 * delta0
       >>> disjoint, _ = parallel_lines_parameters(start0, end0, start1, end1)
       >>> disjoint
       True

    .. testcleanup:: parallel-different2

       import make_images
       make_images.helper_parallel_lines(
           start0, end0, start1, end1, "parallel_lines_parameters2.png")

    but if the intervals overlap, like :math:`\left[0, 1\right]` and
    :math:`\left[-1, \frac{1}{2}\right]`, the segments meet:

    .. image:: ../images/parallel_lines_parameters3.png
       :align: center

    .. testsetup:: parallel-different3, parallel-different4

       import numpy as np
       from bezier._py_geometric_intersection import parallel_lines_parameters

       start0 = np.asfortranarray([1.0, 0.0])
       delta0 = np.asfortranarray([2.0, -1.0])
       end0 = start0 + 1.0 * delta0

    .. doctest:: parallel-different3

       >>> start1 = start0 - 1.5 * delta0
       >>> end1 = start0 + 0.5 * delta0
       >>> disjoint, parameters = parallel_lines_parameters(
       ...     start0, end0, start1, end1)
       >>> disjoint
       False
       >>> parameters
       array([[0.  , 0.5 ],
              [0.75, 1.  ]])

    .. testcleanup:: parallel-different3

       import make_images
       make_images.helper_parallel_lines(
           start0, end0, start1, end1, "parallel_lines_parameters3.png")

    Similarly, if the second interval completely contains the first,
    the segments meet:

    .. image:: ../images/parallel_lines_parameters4.png
       :align: center

    .. doctest:: parallel-different4

       >>> start1 = start0 + 4.5 * delta0
       >>> end1 = start0 - 3.5 * delta0
       >>> disjoint, parameters = parallel_lines_parameters(
       ...     start0, end0, start1, end1)
       >>> disjoint
       False
       >>> parameters
       array([[1.    , 0.    ],
              [0.4375, 0.5625]])

    .. testcleanup:: parallel-different4

       import make_images
       make_images.helper_parallel_lines(
           start0, end0, start1, end1, "parallel_lines_parameters4.png")

    .. note::

       This function doesn't currently allow wiggle room around the
       desired value, i.e. the two values must be bitwise identical.
       However, the most "correct" version of this function likely
       should allow for some round off.

    .. note::

       There is also a Fortran implementation of this function, which
       will be used if it can be built.

    Args:
        start0 (numpy.ndarray): A 1D NumPy ``2``-array that is the start
            vector :math:`S_0` of the parametric line :math:`L_0(s)`.
        end0 (numpy.ndarray): A 1D NumPy ``2``-array that is the end
            vector :math:`E_0` of the parametric line :math:`L_0(s)`.
        start1 (numpy.ndarray): A 1D NumPy ``2``-array that is the start
            vector :math:`S_1` of the parametric line :math:`L_1(s)`.
        end1 (numpy.ndarray): A 1D NumPy ``2``-array that is the end
            vector :math:`E_1` of the parametric line :math:`L_1(s)`.

    Returns:
        Tuple[bool, Optional[numpy.ndarray]]: A pair of

        * Flag indicating if the lines are disjoint.
        * An optional ``2 x 2`` matrix of ``s-t`` parameters only present if
          the lines aren't disjoint. The first column will contain the
          parameters at the beginning of the shared segment and the second
          column will correspond to the end of the shared segment.
    """
    # NOTE: There is no corresponding "enable", but the disable only applies
    #       in this lexical scope.
    # pylint: disable=too-many-branches
    delta0 = end0 - start0
    line0_const = _py_helpers.cross_product(start0, delta0)
    start1_against = _py_helpers.cross_product(start1, delta0)
    if line0_const != start1_against:
        return True, None

    # Each array is a 1D vector, so we can use the vector dot product.
    norm0_sq = np.vdot(delta0, delta0)
    #                S1 = L1(0) = S0 + sA D0
    # <==>        sA D0 = S1 - S0
    #  ==> sA (D0^T D0) = D0^T (S1 - S0)
    s_val0 = np.vdot(start1 - start0, delta0) / norm0_sq
    #                E1 = L1(1) = S0 + sB D0
    # <==>        sB D0 = E1 - S0
    #  ==> sB (D0^T D0) = D0^T (E1 - S0)
    s_val1 = np.vdot(end1 - start0, delta0) / norm0_sq
    # s = s_val0 + t (s_val1 - s_val0)
    # t = 0                                <==> s = s_val0
    # t = 1                                <==> s = s_val1
    # t = -s_val0 / (s_val1 - s_val0)      <==> s = 0
    # t = (1 - s_val0) / (s_val1 - s_val0) <==> s = 1
    if s_val0 <= s_val1:
        # In this branch the segments are moving in the same direction, i.e.
        # (t=0<-->s=s_val0) are both less than (t=1<-->s_val1).
        if 1.0 < s_val0:
            return True, None

        elif s_val0 < 0.0:
            start_s = 0.0
            start_t = -s_val0 / (s_val1 - s_val0)
        else:
            start_s = s_val0
            start_t = 0.0
        if s_val1 < 0.0:
            return True, None

        elif 1.0 < s_val1:
            end_s = 1.0
            end_t = (1.0 - s_val0) / (s_val1 - s_val0)
        else:
            end_s = s_val1
            end_t = 1.0
    else:
        # In this branch the segments are moving in opposite directions, i.e.
        # in (t=0<-->s=s_val0) and (t=1<-->s_val1) we have 0 < 1
        # but ``s_val0 > s_val1``.
        if s_val0 < 0.0:
            return True, None

        elif 1.0 < s_val0:
            start_s = 1.0
            start_t = (s_val0 - 1.0) / (s_val0 - s_val1)
        else:
            start_s = s_val0
            start_t = 0.0
        if 1.0 < s_val1:
            return True, None

        elif s_val1 < 0.0:
            end_s = 0.0
            end_t = s_val0 / (s_val0 - s_val1)
        else:
            end_s = s_val1
            end_t = 1.0
    parameters = np.asfortranarray([[start_s, end_s], [start_t, end_t]])
    return False, parameters


def line_line_collide(line1, line2):
    """Determine if two line segments meet.

    This is a helper for :func:`convex_hull_collide` in the
    special case that the two convex hulls are actually
    just line segments. (Even in this case, this is only
    problematic if both segments are on a single line.)

    Args:
        line1 (numpy.ndarray): ``2 x 2`` array of start and end nodes.
        line2 (numpy.ndarray): ``2 x 2`` array of start and end nodes.

    Returns:
        bool: Indicating if the line segments collide.
    """
    s, t, success = segment_intersection(
        line1[:, 0], line1[:, 1], line2[:, 0], line2[:, 1]
    )
    if success:
        return _py_helpers.in_interval(
            s, 0.0, 1.0
        ) and _py_helpers.in_interval(t, 0.0, 1.0)

    else:
        disjoint, _ = parallel_lines_parameters(
            line1[:, 0], line1[:, 1], line2[:, 0], line2[:, 1]
        )
        return not disjoint


def convex_hull_collide(nodes1, nodes2):
    """Determine if the convex hulls of two curves collide.

    .. note::

       This is a helper for :func:`from_linearized`.

    Args:
        nodes1 (numpy.ndarray): Control points of a first curve.
        nodes2 (numpy.ndarray): Control points of a second curve.

    Returns:
        bool: Indicating if the convex hulls collide.
    """
    polygon1 = _py_helpers.simple_convex_hull(nodes1)
    _, polygon_size1 = polygon1.shape
    polygon2 = _py_helpers.simple_convex_hull(nodes2)
    _, polygon_size2 = polygon2.shape
    if polygon_size1 == 2 and polygon_size2 == 2:
        return line_line_collide(polygon1, polygon2)

    else:
        return _py_helpers.polygon_collide(polygon1, polygon2)


def from_linearized(first, second, intersections):
    """Determine curve-curve intersection from pair of linearizations.

    .. note::

       This assumes that at least one of ``first`` and ``second`` is
       not a line. The line-line case should be handled "early"
       by :func:`check_lines`.

    .. note::

       This assumes the caller has verified that the bounding boxes
       for ``first`` and ``second`` actually intersect.

    If there is an intersection along the segments, adds that intersection
    to ``intersections``. Otherwise, returns without doing anything.

    Args:
        first (Linearization): First curve being intersected.
        second (Linearization): Second curve being intersected.
        intersections (list): A list of existing intersections.

    Raises:
        ValueError: If ``first`` and ``second`` both have linearization error
            of ``0.0`` (i.e. they are both lines). This is because this
            function expects the caller to have used :func:`check_lines`
            already.
    """
    # NOTE: There is no corresponding "enable", but the disable only applies
    #       in this lexical scope.
    # pylint: disable=too-many-return-statements
    s, t, success = segment_intersection(
        first.start_node, first.end_node, second.start_node, second.end_node
    )
    bad_parameters = False
    if success:
        if not (
            _py_helpers.in_interval(s, 0.0, 1.0)
            and _py_helpers.in_interval(t, 0.0, 1.0)
        ):
            bad_parameters = True
    else:
        if first.error == 0.0 and second.error == 0.0:
            raise ValueError(_UNHANDLED_LINES)

        # Just fall back to a Newton iteration starting in the middle of
        # the given intervals.
        bad_parameters = True
        s = 0.5
        t = 0.5
    if bad_parameters:
        # In the unlikely case that we have parallel segments or segments
        # that intersect outside of [0, 1] x [0, 1], we can still exit
        # if the convex hulls don't intersect.
        if not convex_hull_collide(first.curve.nodes, second.curve.nodes):
            return

    # Now, promote ``s`` and ``t`` onto the original curves.
    orig_s = (1 - s) * first.curve.start + s * first.curve.end
    orig_t = (1 - t) * second.curve.start + t * second.curve.end
    refined_s, refined_t = _py_intersection_helpers.full_newton(
        orig_s, first.curve.original_nodes, orig_t, second.curve.original_nodes
    )
    refined_s, success = _py_helpers.wiggle_interval(refined_s)
    if not success:
        return

    refined_t, success = _py_helpers.wiggle_interval(refined_t)
    if not success:
        return

    add_intersection(refined_s, refined_t, intersections)


def add_intersection(s, t, intersections):
    r"""Adds an intersection to list of ``intersections``.

    .. note::

       This is a helper for :func:`from_linearized` and :func:`endpoint_check`.
       These functions are used (directly or indirectly) by
       :func:`all_intersections` exclusively, and that function has a
       Fortran equivalent.

    Accounts for repeated intersection points. If the intersection has already
    been found, does nothing.

    If ``s`` is below :math:`2^{-10}`, it will be replaced with ``1 - s``
    and compared against ``1 - s'`` for all ``s'`` already in
    ``intersections``. (Similar if ``t`` is below the :attr:`.ZERO_THRESHOLD`.)
    This is perfectly "appropriate" since evaluating a B |eacute| zier curve
    requires using both ``s`` and ``1 - s``, so both values are equally
    relevant.

    Compares :math:`\|p - q\|` to :math:`\|p\|` where :math:`p = (s, t)` is
    current candidate intersection (or the "normalized" version, such as
    :math:`p = (1 - s, t)`) and :math:`q` is one of the already added
    intersections. If the difference is below :math:`2^{-42}` (i.e.
    :attr`.NEWTON_ERROR_RATIO`) then the intersection is considered to be
    duplicate.

    Args:
        s (float): The first parameter in an intersection.
        t (float): The second parameter in an intersection.
        intersections (list): List of existing intersections.
    """
    if not intersections:
        intersections.append((s, t))
        return

    if s < _py_intersection_helpers.ZERO_THRESHOLD:
        candidate_s = 1.0 - s
    else:
        candidate_s = s
    if t < _py_intersection_helpers.ZERO_THRESHOLD:
        candidate_t = 1.0 - t
    else:
        candidate_t = t
    norm_candidate = np.linalg.norm([candidate_s, candidate_t], ord=2)
    for existing_s, existing_t in intersections:
        # NOTE: |(1 - s1) - (1 - s2)| = |s1 - s2| in exact arithmetic, so
        #       we just compute ``s1 - s2`` rather than using
        #       ``candidate_s`` / ``candidate_t``. Due to round-off, these
        #       differences may be slightly different, but only up to machine
        #       precision.
        delta_s = s - existing_s
        delta_t = t - existing_t
        norm_update = np.linalg.norm([delta_s, delta_t], ord=2)
        if (
            norm_update
            < _py_intersection_helpers.NEWTON_ERROR_RATIO * norm_candidate
        ):
            return

    intersections.append((s, t))


def endpoint_check(
    first, node_first, s, second, node_second, t, intersections
):
    r"""Check if curve endpoints are identical.

    .. note::

       This is a helper for :func:`tangent_bbox_intersection`. These
       functions are used (directly or indirectly) by
       :func:`all_intersections` exclusively, and that function has a
       Fortran equivalent.

    Args:
        first (SubdividedCurve): First curve being intersected (assumed in
            :math:\mathbf{R}^2`).
        node_first (numpy.ndarray): 1D ``2``-array, one of the endpoints
            of ``first``.
        s (float): The parameter corresponding to ``node_first``, so
             expected to be one of ``0.0`` or ``1.0``.
        second (SubdividedCurve): Second curve being intersected (assumed in
            :math:\mathbf{R}^2`).
        node_second (numpy.ndarray): 1D ``2``-array, one of the endpoints
            of ``second``.
        t (float): The parameter corresponding to ``node_second``, so
             expected to be one of ``0.0`` or ``1.0``.
        intersections (list): A list of already encountered
            intersections. If these curves intersect at their tangency,
            then those intersections will be added to this list.
    """
    if _py_helpers.vector_close(node_first, node_second):
        orig_s = (1 - s) * first.start + s * first.end
        orig_t = (1 - t) * second.start + t * second.end
        add_intersection(orig_s, orig_t, intersections)


def tangent_bbox_intersection(first, second, intersections):
    r"""Check if two curves with tangent bounding boxes intersect.

    .. note::

       This is a helper for :func:`intersect_one_round`. These
       functions are used (directly or indirectly) by
       :func:`all_intersections` exclusively, and that function has a
       Fortran equivalent.

    If the bounding boxes are tangent, intersection can
    only occur along that tangency.

    If the curve is **not** a line, the **only** way the curve can touch
    the bounding box is at the endpoints. To see this, consider the
    component

    .. math::

       x(s) = \sum_j W_j x_j.

    Since :math:`W_j > 0` for :math:`s \in \left(0, 1\right)`, if there
    is some :math:`k` with :math:`x_k < M = \max x_j`, then for any
    interior :math:`s`

    .. math::

       x(s) < \sum_j W_j M = M.

    If all :math:`x_j = M`, then :math:`B(s)` falls on the line
    :math:`x = M`. (A similar argument holds for the other three
    component-extrema types.)

    .. note::

       This function assumes callers will not pass curves that can be
       linearized / are linear. In :func:`all_intersections`, curves
       are pre-processed to do any linearization before the
       subdivision / intersection process begins.

    Args:
        first (SubdividedCurve): First curve being intersected (assumed in
            :math:\mathbf{R}^2`).
        second (SubdividedCurve): Second curve being intersected (assumed in
            :math:\mathbf{R}^2`).
        intersections (list): A list of already encountered
            intersections. If these curves intersect at their tangency,
            then those intersections will be added to this list.
    """
    node_first1 = first.nodes[:, 0]
    node_first2 = first.nodes[:, -1]
    node_second1 = second.nodes[:, 0]
    node_second2 = second.nodes[:, -1]
    endpoint_check(
        first, node_first1, 0.0, second, node_second1, 0.0, intersections
    )
    endpoint_check(
        first, node_first1, 0.0, second, node_second2, 1.0, intersections
    )
    endpoint_check(
        first, node_first2, 1.0, second, node_second1, 0.0, intersections
    )
    endpoint_check(
        first, node_first2, 1.0, second, node_second2, 1.0, intersections
    )


def bbox_line_intersect(nodes, line_start, line_end):
    r"""Determine intersection of a bounding box and a line.

    We do this by first checking if either the start or end node of the
    segment are contained in the bounding box. If they aren't, then
    checks if the line segment intersects any of the four sides of the
    bounding box.

    .. note::

       This function is "half-finished". It makes no distinction between
       "tangent" intersections of the box and segment and other types
       of intersection. However, the distinction is worthwhile, so this
       function should be "upgraded" at some point.

    Args:
        nodes (numpy.ndarray): Points (``2 x N``) that determine a
            bounding box.
        line_start (numpy.ndarray): Beginning of a line segment (1D
            ``2``-array).
        line_end (numpy.ndarray): End of a line segment (1D ``2``-array).

    Returns:
        int: Enum from ``BoxIntersectionType`` indicating the type of
        bounding box intersection.
    """
    left, right, bottom, top = _py_helpers.bbox(nodes)
    if _py_helpers.in_interval(
        line_start[0], left, right
    ) and _py_helpers.in_interval(line_start[1], bottom, top):
        return BoxIntersectionType.INTERSECTION

    if _py_helpers.in_interval(
        line_end[0], left, right
    ) and _py_helpers.in_interval(line_end[1], bottom, top):
        return BoxIntersectionType.INTERSECTION

    # NOTE: We allow ``segment_intersection`` to fail below (i.e.
    #       ``success=False``). At first, this may appear to "ignore"
    #       some potential intersections of parallel lines. However,
    #       no intersections will be missed. If parallel lines don't
    #       overlap, then there is nothing to miss. If they do overlap,
    #       then either the segment will have endpoints on the box (already
    #       covered by the checks above) or the segment will contain an
    #       entire side of the box, which will force it to intersect the 3
    #       edges that meet at the two ends of those sides. The parallel
    #       edge will be skipped, but the other two will be covered.
    # Bottom Edge
    s_bottom, t_bottom, success = segment_intersection(
        np.asfortranarray([left, bottom]),
        np.asfortranarray([right, bottom]),
        line_start,
        line_end,
    )
    if (
        success
        and _py_helpers.in_interval(s_bottom, 0.0, 1.0)
        and _py_helpers.in_interval(t_bottom, 0.0, 1.0)
    ):
        return BoxIntersectionType.INTERSECTION

    # Right Edge
    s_right, t_right, success = segment_intersection(
        np.asfortranarray([right, bottom]),
        np.asfortranarray([right, top]),
        line_start,
        line_end,
    )
    if (
        success
        and _py_helpers.in_interval(s_right, 0.0, 1.0)
        and _py_helpers.in_interval(t_right, 0.0, 1.0)
    ):
        return BoxIntersectionType.INTERSECTION

    # Top Edge
    s_top, t_top, success = segment_intersection(
        np.asfortranarray([right, top]),
        np.asfortranarray([left, top]),
        line_start,
        line_end,
    )
    if (
        success
        and _py_helpers.in_interval(s_top, 0.0, 1.0)
        and _py_helpers.in_interval(t_top, 0.0, 1.0)
    ):
        return BoxIntersectionType.INTERSECTION

    # NOTE: We skip the "last" edge. This is because any curve
    #       that doesn't have an endpoint on a curve must cross
    #       at least two, so we will have already covered such curves
    #       in one of the branches above.
    return BoxIntersectionType.DISJOINT


def intersect_one_round(candidates, intersections):
    """Perform one step of the intersection process.

    .. note::

       This is a helper for :func:`all_intersections` and that function
       has a Fortran equivalent.

    Checks if the bounding boxes of each pair in ``candidates``
    intersect. If the bounding boxes do not intersect, the pair
    is discarded. Otherwise, the pair is "accepted". Then we
    attempt to linearize each curve in an "accepted" pair and
    track the overall linearization error for every curve
    encountered.

    Args:
        candidates (Union[list, itertools.chain]): An iterable of
            pairs of curves (or linearized curves).
        intersections (list): A list of already encountered
            intersections. If any intersections can be readily determined
            during this round of subdivision, then they will be added
            to this list.

    Returns:
        list: Returns a list of the next round of ``candidates``.
    """
    next_candidates = []
    # NOTE: In the below we replace ``isinstance(a, B)`` with
    #       ``a.__class__ is B``, which is a 3-3.5x speedup.
    for first, second in candidates:
        both_linearized = False
        if first.__class__ is Linearization:
            if second.__class__ is Linearization:
                both_linearized = True
                bbox_int = bbox_intersect(
                    first.curve.nodes, second.curve.nodes
                )
            else:
                bbox_int = bbox_line_intersect(
                    second.nodes, first.start_node, first.end_node
                )
        else:
            if second.__class__ is Linearization:
                bbox_int = bbox_line_intersect(
                    first.nodes, second.start_node, second.end_node
                )
            else:
                bbox_int = bbox_intersect(first.nodes, second.nodes)
        if bbox_int == BoxIntersectionType.DISJOINT:
            continue

        if bbox_int == BoxIntersectionType.TANGENT and not both_linearized:
            # NOTE: Ignore tangent bounding boxes in the linearized case
            #       because ``tangent_bbox_intersection()`` assumes that both
            #       curves are not linear.
            tangent_bbox_intersection(first, second, intersections)
            continue

        if both_linearized:
            # If both ``first`` and ``second`` are linearizations, then
            # we can intersect them immediately.
            from_linearized(first, second, intersections)
            continue

        # If we haven't ``continue``-d, add the accepted pair.
        # NOTE: This may be a wasted computation, e.g. if ``first``
        #       or ``second`` occur in multiple accepted pairs (the caller
        #       only passes one pair at a time). However, in practice
        #       the number of such pairs will be small so this cost
        #       will be low.
        lin1 = map(Linearization.from_shape, first.subdivide())
        lin2 = map(Linearization.from_shape, second.subdivide())
        next_candidates.extend(itertools.product(lin1, lin2))
    return next_candidates


def prune_candidates(candidates):
    """Reduce number of candidate intersection pairs.

    .. note::

       This is a helper for :func:`all_intersections`.

    Uses more strict bounding box intersection predicate by forming the
    actual convex hull of each candidate curve segment and then checking
    if those convex hulls collide.

    Args:
        candidates (List): An iterable of pairs of curves (or
            linearized curves).

    Returns:
        List: A pruned list of curve pairs.
    """
    pruned = []
    # NOTE: In the below we replace ``isinstance(a, B)`` with
    #       ``a.__class__ is B``, which is a 3-3.5x speedup.
    for first, second in candidates:
        if first.__class__ is Linearization:
            nodes1 = first.curve.nodes
        else:
            nodes1 = first.nodes
        if second.__class__ is Linearization:
            nodes2 = second.curve.nodes
        else:
            nodes2 = second.nodes
        if convex_hull_collide(nodes1, nodes2):
            pruned.append((first, second))
    return pruned


def make_same_degree(nodes1, nodes2):
    """Degree-elevate a curve so two curves have matching degree.

    Args:
        nodes1 (numpy.ndarray): Set of control points for a
            B |eacute| zier curve.
        nodes2 (numpy.ndarray): Set of control points for a
            B |eacute| zier curve.

    Returns:
        Tuple[numpy.ndarray, numpy.ndarray]: The potentially degree-elevated
        nodes passed in.
    """
    _, num_nodes1 = nodes1.shape
    _, num_nodes2 = nodes2.shape
    for _ in range(num_nodes2 - num_nodes1):
        nodes1 = _py_curve_helpers.elevate_nodes(nodes1)
    for _ in range(num_nodes1 - num_nodes2):
        nodes2 = _py_curve_helpers.elevate_nodes(nodes2)
    return nodes1, nodes2


def coincident_parameters(nodes1, nodes2):
    r"""Check if two B |eacute| zier curves are coincident.

    Does so by projecting each segment endpoint onto the other curve

    .. math::

       B_1(s_0) = B_2(0)
       B_1(s_m) = B_2(1)
       B_1(0) = B_2(t_0)
       B_1(1) = B_2(t_n)

    and then finding the "shared interval" where both curves are defined.
    If such an interval can't be found (e.g. if one of the endpoints can't be
    located on the other curve), returns :data:`None`.

    If such a "shared interval" does exist, then this will specialize
    each curve onto that shared interval and check if the new control points
    agree.

    Args:
        nodes1 (numpy.ndarray): Set of control points for a
            B |eacute| zier curve.
        nodes2 (numpy.ndarray): Set of control points for a
            B |eacute| zier curve.

    Returns:
        Optional[Tuple[Tuple[float, float], ...]]: A ``2 x 2`` array of
        parameters where the two coincident curves meet. If they are not
        coincident, returns :data:`None`.
    """
    # NOTE: There is no corresponding "enable", but the disable only applies
    #       in this lexical scope.
    # pylint: disable=too-many-return-statements,too-many-branches
    nodes1, nodes2 = make_same_degree(nodes1, nodes2)
    s_initial = _py_curve_helpers.locate_point(
        nodes1, nodes2[:, 0].reshape((2, 1), order="F")
    )
    s_final = _py_curve_helpers.locate_point(
        nodes1, nodes2[:, -1].reshape((2, 1), order="F")
    )
    if s_initial is not None and s_final is not None:
        # In this case, if the curves were coincident, then ``curve2``
        # would be "fully" contained in ``curve1``, so we specialize
        # ``curve1`` down to that interval to check.
        specialized1 = _py_curve_helpers.specialize_curve(
            nodes1, s_initial, s_final
        )
        if _py_helpers.vector_close(
            specialized1.ravel(order="F"), nodes2.ravel(order="F")
        ):
            return ((s_initial, 0.0), (s_final, 1.0))

        else:
            return None

    t_initial = _py_curve_helpers.locate_point(
        nodes2, nodes1[:, 0].reshape((2, 1), order="F")
    )
    t_final = _py_curve_helpers.locate_point(
        nodes2, nodes1[:, -1].reshape((2, 1), order="F")
    )
    if t_initial is None and t_final is None:
        # An overlap must have two endpoints and since at most one of the
        # endpoints of ``curve2`` lies on ``curve1`` (as indicated by at
        # least one of the ``s``-parameters being ``None``), we need (at least)
        # one endpoint of ``curve1`` on ``curve2``.
        return None

    if t_initial is not None and t_final is not None:
        # In this case, if the curves were coincident, then ``curve1``
        # would be "fully" contained in ``curve2``, so we specialize
        # ``curve2`` down to that interval to check.
        specialized2 = _py_curve_helpers.specialize_curve(
            nodes2, t_initial, t_final
        )
        if _py_helpers.vector_close(
            nodes1.ravel(order="F"), specialized2.ravel(order="F")
        ):
            return ((0.0, t_initial), (1.0, t_final))

        else:
            return None

    if s_initial is None and s_final is None:
        # An overlap must have two endpoints and since exactly one of the
        # endpoints of ``curve1`` lies on ``curve2`` (as indicated by exactly
        # one of the ``t``-parameters being ``None``), we need (at least)
        # one endpoint of ``curve1`` on ``curve2``.
        return None

    # At this point, we know exactly one of the ``s``-parameters and exactly
    # one of the ``t``-parameters is not ``None``.
    if s_initial is None:
        if t_initial is None:
            # B1(s_final) = B2(1) AND B1(1) = B2(t_final)
            start_s = s_final
            end_s = 1.0
            start_t = 1.0
            end_t = t_final
        else:
            # B1(0) = B2(t_initial) AND B1(s_final) = B2(1)
            start_s = 0.0
            end_s = s_final
            start_t = t_initial
            end_t = 1.0
    else:
        if t_initial is None:
            # B1(s_initial) = B2(0) AND B1(1 ) = B2(t_final)
            start_s = s_initial
            end_s = 1.0
            start_t = 0.0
            end_t = t_final
        else:
            # B1(0) = B2(t_initial) AND B1(s_initial) = B2(0)
            start_s = 0.0
            end_s = s_initial
            start_t = t_initial
            end_t = 0.0
    width_s = abs(start_s - end_s)
    width_t = abs(start_t - end_t)
    if width_s < _MIN_INTERVAL_WIDTH and width_t < _MIN_INTERVAL_WIDTH:
        return None

    specialized1 = _py_curve_helpers.specialize_curve(nodes1, start_s, end_s)
    specialized2 = _py_curve_helpers.specialize_curve(nodes2, start_t, end_t)
    if _py_helpers.vector_close(
        specialized1.ravel(order="F"), specialized2.ravel(order="F")
    ):
        return ((start_s, start_t), (end_s, end_t))

    else:
        return None


def check_lines(first, second):
    """Checks if two curves are lines and tries to intersect them.

    .. note::

       This is a helper for :func:`.all_intersections`.

    If they are not lines / not linearized, immediately returns :data:`False`
    with no "return value".

    If they are lines, attempts to intersect them (even if they are parallel
    and share a coincident segment).

    Args:
        first (Union[SubdividedCurve, Linearization]): First curve being
            intersected.
        second (Union[SubdividedCurve, Linearization]): Second curve being
            intersected.

    Returns:
        Tuple[bool, Optional[Tuple[numpy.ndarray, bool]]]: A pair of

        * Flag indicating if both candidates in the pair are lines.
        * Optional "result" populated only if both candidates are lines.
          When this result is populated, it will be a pair of

          * array of parameters of intersection
          * flag indicating if the two candidates share a coincident segment
    """
    # NOTE: In the below we replace ``isinstance(a, B)`` with
    #       ``a.__class__ is B``, which is a 3-3.5x speedup.
    if not (
        first.__class__ is Linearization
        and second.__class__ is Linearization
        and first.error == 0.0
        and second.error == 0.0
    ):
        return False, None

    s, t, success = segment_intersection(
        first.start_node, first.end_node, second.start_node, second.end_node
    )
    if success:
        if _py_helpers.in_interval(s, 0.0, 1.0) and _py_helpers.in_interval(
            t, 0.0, 1.0
        ):
            intersections = np.asfortranarray([[s], [t]])
            result = intersections, False
        else:
            result = np.empty((2, 0), order="F"), False
    else:
        disjoint, params = parallel_lines_parameters(
            first.start_node,
            first.end_node,
            second.start_node,
            second.end_node,
        )
        if disjoint:
            result = np.empty((2, 0), order="F"), False
        else:
            result = params, True
    return True, result


def all_intersections(nodes_first, nodes_second):
    r"""Find the points of intersection among a pair of curves.

    .. note::

       There is also a Fortran implementation of this function, which
       will be used if it can be built.

    .. note::

       This assumes both curves are in :math:`\mathbf{R}^2`, but does not
       **explicitly** check this. However, functions used here will fail if
       that assumption fails, e.g. :func:`bbox_intersect` and
       :func:`newton_refine() <._py_intersection_helpers.newton_refine>`.

    Args:
        nodes_first (numpy.ndarray): Control points of a curve to be
            intersected with ``nodes_second``.
        nodes_second (numpy.ndarray): Control points of a curve to be
            intersected with ``nodes_first``.

    Returns:
        Tuple[numpy.ndarray, bool]: An array and a flag:

        * A ``2 x N`` array of intersection parameters.
          Each row contains a pair of values :math:`s` and :math:`t`
          (each in :math:`\left[0, 1\right]`) such that the curves
          intersect: :math:`B_1(s) = B_2(t)`.
        * Flag indicating if the curves are coincident.

    Raises:
        ValueError: If the subdivision iteration does not terminate
            before exhausting the maximum number of subdivisions.
        NotImplementedError: If the subdivision process picks up too
            many candidate pairs. This typically indicates tangent
            curves or coincident curves (though there are mitigations for
            those cases in place).
    """
    curve_first = SubdividedCurve(nodes_first, nodes_first)
    curve_second = SubdividedCurve(nodes_second, nodes_second)
    candidate1 = Linearization.from_shape(curve_first)
    candidate2 = Linearization.from_shape(curve_second)
    # Handle the line-line intersection case as a one-off.
    both_linear, result = check_lines(candidate1, candidate2)
    if both_linear:
        return result

    candidates = [(candidate1, candidate2)]
    intersections = []
    coincident = False
    for _ in range(_MAX_INTERSECT_SUBDIVISIONS):
        candidates = intersect_one_round(candidates, intersections)
        if len(candidates) > _MAX_CANDIDATES:
            candidates = prune_candidates(candidates)
            # If pruning didn't fix anything, we check if the curves are
            # coincident and "fail" if they aren't.
            if len(candidates) > _MAX_CANDIDATES:
                params = coincident_parameters(nodes_first, nodes_second)
                if params is None:
                    raise NotImplementedError(
                        _TOO_MANY_TEMPLATE.format(len(candidates))
                    )

                intersections = params
                coincident = True
                # Artificially empty out candidates so that this
                # function exits.
                candidates = []
        # If none of the candidate pairs have been accepted, then there are
        # no more intersections to find.
        if not candidates:
            if intersections:
                # NOTE: The transpose of a C-ordered array is Fortran-ordered,
                #       i.e. this is on purpose.
                return np.array(intersections, order="C").T, coincident

            return np.empty((2, 0), order="F"), coincident

    msg = _NO_CONVERGE_TEMPLATE.format(_MAX_INTERSECT_SUBDIVISIONS)
    raise ValueError(msg)


class BoxIntersectionType:  # pylint: disable=too-few-public-methods
    """Enum representing all possible bounding box intersections.

    .. note::

       This class would be more "correct" as an ``enum.Enum``, but it we keep
       the values integers to make interfacing with Fortran easier.
    """

    INTERSECTION = 0
    """Bounding boxes overlap with positive area."""
    TANGENT = 1
    """Bounding boxes are tangent but do not overlap."""
    DISJOINT = 2
    """Bounding boxes do not intersect or touch."""


class SubdividedCurve:  # pylint: disable=too-few-public-methods
    """A data wrapper for a B |eacute| zier curve

    To be used for intersection algorithm via repeated subdivision,
    where the ``start`` and ``end`` parameters must be tracked.

    Args:
        nodes (numpy.ndarray): The control points of the current
            subdivided curve
        original_nodes (numpy.ndarray): The control points of the original
            curve used to define the current one (before subdivision began).
        start (Optional[float]): The start parameter after subdivision.
        end (Optional[float]): The start parameter after subdivision.
    """

    __slots__ = ("nodes", "original_nodes", "start", "end")

    def __init__(self, nodes, original_nodes, start=0.0, end=1.0):
        self.nodes = nodes
        self.original_nodes = original_nodes
        self.start = start
        self.end = end

    @property
    def __dict__(self):
        """dict: Dictionary of current subdivided curve's property namespace.

        This is just a stand-in property for the usual ``__dict__``. This
        class defines ``__slots__`` so by default would not provide a
        ``__dict__``.

        This also means that the current object can't be modified by the
        returned dictionary.
        """
        return {
            "nodes": self.nodes,
            "original_nodes": self.original_nodes,
            "start": self.start,
            "end": self.end,
        }

    def subdivide(self):
        """Split the curve into a left and right half.

        See :meth:`.Curve.subdivide` for more information.

        Returns:
            Tuple[SubdividedCurve, SubdividedCurve]: The left and right
            sub-curves.
        """
        left_nodes, right_nodes = _py_curve_helpers.subdivide_nodes(self.nodes)
        midpoint = 0.5 * (self.start + self.end)
        left = SubdividedCurve(
            left_nodes, self.original_nodes, start=self.start, end=midpoint
        )
        right = SubdividedCurve(
            right_nodes, self.original_nodes, start=midpoint, end=self.end
        )
        return left, right


class Linearization:
    """A linearization of a curve.

    This class is provided as a stand-in for a curve, so it
    provides a similar interface.

    Args:
        curve (SubdividedCurve): A curve that is linearized.
        error (float): The linearization error. Expected to have been
            computed via :func:`linearization_error`.
    """

    __slots__ = ("curve", "error", "start_node", "end_node")

    def __init__(self, curve, error):
        self.curve = curve
        """SubdividedCurve: The curve that this linearization approximates."""
        self.error = error
        """float: The linearization error for the linearized curve."""
        self.start_node = curve.nodes[:, 0]
        """numpy.ndarray: The 1D start vector of this linearization."""
        self.end_node = curve.nodes[:, -1]
        """numpy.ndarray: The 1D end vector of this linearization."""

    @property
    def __dict__(self):
        """dict: Dictionary of current linearization's property namespace.

        This is just a stand-in property for the usual ``__dict__``. This
        class defines ``__slots__`` so by default would not provide a
        ``__dict__``.

        This also means that the current object can't be modified by the
        returned dictionary.
        """
        return {
            "curve": self.curve,
            "error": self.error,
            "start_node": self.start_node,
            "end_node": self.end_node,
        }

    def subdivide(self):
        """Do-nothing method to match the :class:`.Curve` interface.

        Returns:
            Tuple[~bezier._py_geometric_intersection.Linearization]: List of
            all subdivided parts, which is just the current object.
        """
        return (self,)

    @classmethod
    def from_shape(cls, shape):
        """Try to linearize a curve (or an already linearized curve).

        Args:
            shape (Union[SubdividedCurve, \
            ~bezier._py_geometric_intersection.Linearization]): A curve or an
                already linearized curve.

        Returns:
            Union[SubdividedCurve, \
            ~bezier._py_geometric_intersection.Linearization]: The
            (potentially linearized) curve.
        """
        # NOTE: In the below we replace ``isinstance(a, B)`` with
        #       ``a.__class__ is B``, which is a 3-3.5x speedup.
        if shape.__class__ is cls:
            return shape

        else:
            error = linearization_error(shape.nodes)
            if error < _ERROR_VAL:
                linearized = cls(shape, error)
                return linearized

            else:
                return shape



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Helpers for ``setup.py`` specific to macOS."""

import sys

import setup_helpers


MAC_OS = "darwin"


def is_macos_gfortran(f90_compiler):
    """Checks if the current build is ``gfortran`` on macOS.

    Args:
        f90_compiler (numpy.distutils.fcompiler.FCompiler): A Fortran compiler
            instance.

    Returns:
        bool: Only :data:`True` if

        * Current OS is macOS (checked via ``sys.platform``).
        * ``f90_compiler`` corresponds to ``gfortran``.
    """
    # NOTE: NumPy may not be installed, but we don't want **this** module to
    #       cause an import failure.
    from numpy.distutils.fcompiler import gnu

    # Only macOS.
    if sys.platform != MAC_OS:
        return False

    # Only ``gfortran``.
    if not isinstance(f90_compiler, gnu.Gnu95FCompiler):
        return False

    return True


def patch_f90_compiler(f90_compiler):
    """Patch up ``f90_compiler.library_dirs``.

    On macOS, a Homebrew installed ``gfortran`` needs some help. The
    ``numpy.distutils`` "default" constructor for ``Gnu95FCompiler`` only has
    a single library search path, but there are many library paths included in
    the full ``gcc`` install.

    Args:
        f90_compiler (numpy.distutils.fcompiler.FCompiler): A Fortran compiler
            instance.
    """
    if not is_macos_gfortran(f90_compiler):
        return

    library_dirs = f90_compiler.library_dirs
    # ``library_dirs`` is a list (i.e. mutable), so we can update in place.
    library_dirs[:] = setup_helpers.gfortran_search_path(library_dirs)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Helper methods for intersecting B |eacute| zier shapes.

The functions provided by this module have a Cython speedup with the
exact same interface which calls out to a Fortran implementation. The speedup
will be used if the extension can be built.
"""

from bezier import _py_intersection_helpers

try:
    from bezier import _speedup
except ImportError:  # pragma: NO COVER
    _speedup = None


# pylint: disable=invalid-name
if _speedup is None:  # pragma: NO COVER
    newton_refine = _py_intersection_helpers.newton_refine
else:
    newton_refine = _speedup.newton_refine_curve_intersect
# pylint: enable=invalid-name



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Helper methods for :mod:`bezier.triangle`.

The functions provided by this module have a Cython speedup with the
exact same interface which calls out to a Fortran implementation. The speedup
will be used if the extension can be built.
"""

from bezier import _py_triangle_helpers

try:
    from bezier import _speedup
except ImportError:  # pragma: NO COVER
    _speedup = None


# pylint: disable=invalid-name
if _speedup is None:  # pragma: NO COVER
    de_casteljau_one_round = _py_triangle_helpers.de_casteljau_one_round
    specialize_triangle = _py_triangle_helpers.specialize_triangle
    subdivide_nodes = _py_triangle_helpers.subdivide_nodes
    jacobian_both = _py_triangle_helpers.jacobian_both
    jacobian_det = _py_triangle_helpers.jacobian_det
    evaluate_barycentric = _py_triangle_helpers.evaluate_barycentric
    evaluate_barycentric_multi = (
        _py_triangle_helpers.evaluate_barycentric_multi
    )
    evaluate_cartesian_multi = _py_triangle_helpers.evaluate_cartesian_multi
    compute_edge_nodes = _py_triangle_helpers.compute_edge_nodes
    compute_area = _py_triangle_helpers.compute_area
else:
    de_casteljau_one_round = _speedup.de_casteljau_one_round
    specialize_triangle = _speedup.specialize_triangle
    subdivide_nodes = _speedup.subdivide_nodes_triangle
    jacobian_both = _speedup.jacobian_both
    jacobian_det = _speedup.jacobian_det
    evaluate_barycentric = _speedup.evaluate_barycentric
    evaluate_barycentric_multi = _speedup.evaluate_barycentric_multi
    evaluate_cartesian_multi = _speedup.evaluate_cartesian_multi
    compute_edge_nodes = _speedup.compute_edge_nodes
    compute_area = _speedup.compute_area
# pylint: enable=invalid-name



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Environment configuration for ``bezier`` runtime dependencies.

Only needed for Windows, to add ``extra-dll`` directory to the DLL search
path so that the ``libbezier`` DLL can be located.
"""

import os

import pkg_resources


# Error messages for ``handle_import_error``.
TEMPLATE = "No module named 'bezier.{}'"  # 3.6, 3.7, 3.8, pypy3
# NOTE: ``os.add_dll_directory()`` was added on Windows in Python 3.8.
OS_ADD_DLL_DIRECTORY = getattr(os, "add_dll_directory", None)


def add_dll_directory(extra_dll_dir):
    """Add a DLL directory.

    This is only expected to be invoked on Windows. For Python versions before
    3.8, this will update the ``%PATH%`` environment variable to include
    ``extra_dll_dir`` and for 3.8 and later, it will invoke
    ``os.add_dll_directory()``.
    """
    if not os.path.isdir(extra_dll_dir):
        return

    if OS_ADD_DLL_DIRECTORY is not None:
        OS_ADD_DLL_DIRECTORY(extra_dll_dir)  # pylint: disable=not-callable
        return

    path = os.environ.get("PATH", "")
    values = [subdir for subdir in path.split(os.pathsep) if subdir]
    values.append(extra_dll_dir)
    os.environ["PATH"] = os.pathsep.join(values)


def modify_path():
    """Add the DLL directory to the module search path.

    This will only modify path if
    * on Windows
    * the ``extra-dll`` directory is in package resources
    """
    if os.name != "nt":
        return

    try:
        extra_dll_dir = pkg_resources.resource_filename("bezier", "extra-dll")
    except ImportError:
        return

    add_dll_directory(extra_dll_dir)


def handle_import_error(caught_exc, name):
    """Allow or re-raise an import error.

    This is to distinguish between expected and unexpected import errors.
    If the module is not found, it simply means the Cython / Fortran speedups
    were not built with the package. If the error message is different, e.g.
    ``... undefined symbol: __curve_intersection_MOD_all_intersections``, then
    the import error **should** be raised.

    Args:
        caught_exc (ImportError): An exception caught when trying to import
            a Cython module.
        name (str): The name of the module. For example, for the module
            ``bezier._curve_speedup``, the name is ``"_curve_speedup"``.

    Raises:
        ImportError: If the error message is different than the basic
            "missing module" error message.
    """
    expected_msg = TEMPLATE.format(name)
    if caught_exc.args == (expected_msg,):
        return

    raise caught_exc


modify_path()



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Cython wrapper for ``helpers.f90``."""


from libcpp cimport bool as bool_t


cdef extern from "bezier/helpers.h":
    void cross_product "BEZ_cross_product" (
        const double* vec0, const double* vec1, double* result)
    void bbox "BEZ_bbox" (
        const int* num_nodes, const double* nodes, double* left,
        double* right, double* bottom, double* top)
    void wiggle_interval "BEZ_wiggle_interval" (
        const double* value, double* result, bool_t* success)
    void contains_nd "BEZ_contains_nd" (
        const int* num_nodes, const int* dimension,
        const double* nodes, const double* point, bool_t* predicate)
    bool_t vector_close "BEZ_vector_close" (
        int* num_values, const double* vec1, const double* vec2,
        const double* eps)
    bool_t in_interval "BEZ_in_interval" (
        const double* value, const double* start, const double* end)
    void simple_convex_hull "BEZ_simple_convex_hull" (
        int* num_points, const double* points, int* polygon_size, double* polygon)
    void polygon_collide "BEZ_polygon_collide" (
        const int* polygon_size1, const double* polygon1,
        const int* polygon_size2, const double* polygon2, bool_t* collision)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Cython wrapper for ``triangle.f90``."""


from libcpp cimport bool as bool_t


cdef extern from "bezier/triangle.h":
    void de_casteljau_one_round "BEZ_de_casteljau_one_round" (
        const int* num_nodes, const int* dimension,
        const double* nodes, const int* degree, const double* lambda1,
        const double* lambda2, const double* lambda3, double* new_nodes)
    void evaluate_barycentric "BEZ_evaluate_barycentric" (
        const int* num_nodes, const int* dimension,
        const double* nodes, const int* degree, const double* lambda1,
        const double* lambda2, const double* lambda3, double* point)
    void evaluate_barycentric_multi "BEZ_evaluate_barycentric_multi" (
        const int* num_nodes, const int* dimension,
        const double* nodes, const int* degree, const int* num_vals,
        const double* param_vals, double* evaluated)
    void evaluate_cartesian_multi "BEZ_evaluate_cartesian_multi" (
        const int* num_nodes, const int* dimension,
        const double* nodes, const int* degree, const int* num_vals,
        const double* param_vals, double* evaluated)
    void jacobian_both "BEZ_jacobian_both" (
        const int* num_nodes, const int* dimension,
        const double* nodes, const int* degree, double* new_nodes)
    void jacobian_det "BEZ_jacobian_det" (
        const int* num_nodes, const double* nodes,
        const int* degree, const int* num_vals, const double* param_vals,
        double* evaluated)
    void specialize_triangle "BEZ_specialize_triangle" (
        const int* num_nodes, const int* dimension,
        const double* nodes, const int* degree, const double* weights_a,
        const double* weights_b, const double* weights_c, double* specialized)
    void subdivide_nodes_triangle "BEZ_subdivide_nodes_triangle" (
        const int* num_nodes, const int* dimension,
        const double* nodes, const int* degree, double* nodes_a, double* nodes_b,
        double* nodes_c, double* nodes_d)
    void compute_edge_nodes "BEZ_compute_edge_nodes" (
        const int* num_nodes, const int* dimension,
        const double* nodes, const int* degree, double* nodes1, double* nodes2,
        double* nodes3)
    void compute_area "BEZ_compute_area" (
        int* num_edges, int* sizes,
        const double* const* nodes_pointers, double* area, bool_t* not_implemented)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Helper for B |eacute| zier Curves.

.. |eacute| unicode:: U+000E9 .. LATIN SMALL LETTER E WITH ACUTE
   :trim:

See :doc:`../../algorithms/curve-curve-intersection` for examples using the
:class:`Curve` class to find intersections.

.. testsetup:: *

   import numpy as np
   import bezier

.. autoclass:: IntersectionStrategy
   :members:
"""

import numpy as np

from bezier import _algebraic_intersection
from bezier import _base
from bezier import _curve_helpers
from bezier import _geometric_intersection
from bezier import _plot_helpers
from bezier import _py_intersection_helpers
from bezier import _symbolic


_LOCATE_ERROR_TEMPLATE = (
    "Dimension mismatch: This curve is {:d}-dimensional, so the point should "
    "be a {:d} x 1 NumPy array. Instead the point {} has dimensions {}."
)
IntersectionStrategy = _py_intersection_helpers.IntersectionStrategy


class Curve(_base.Base):
    r"""Represents a B |eacute| zier `curve`_.

    .. _curve: https://en.wikipedia.org/wiki/B%C3%A9zier_curve

    We take the traditional definition: a B |eacute| zier curve is a mapping
    from :math:`s \in \left[0, 1\right]` to convex combinations
    of points :math:`v_0, v_1, \ldots, v_n` in some vector space:

    .. math::

       B(s) = \sum_{j = 0}^n \binom{n}{j} s^j (1 - s)^{n - j} \cdot v_j

    .. image:: ../../images/curve_constructor.png
       :align: center

    .. doctest:: curve-constructor

       >>> import bezier
       >>> nodes = np.asfortranarray([
       ...     [0.0, 0.625, 1.0],
       ...     [0.0, 0.5  , 0.5],
       ... ])
       >>> curve = bezier.Curve(nodes, degree=2)
       >>> curve
       <Curve (degree=2, dimension=2)>

    .. testcleanup:: curve-constructor

       import make_images
       make_images.curve_constructor(curve)

    Args:
        nodes (Sequence[Sequence[numbers.Number]]): The nodes in the curve.
            Must be convertible to a 2D NumPy array of floating point values,
            where the columns represent each node while the rows are the
            dimension of the ambient space.
        degree (int): The degree of the curve. This is assumed to
            correctly correspond to the number of ``nodes``. Use
            :meth:`from_nodes` if the degree has not yet been computed.
        copy (bool): Flag indicating if the nodes should be copied before
            being stored. Defaults to :data:`True` since callers may
            freely mutate ``nodes`` after passing in.
        verify (bool): Flag indicating if the degree should be verified against
            the number of nodes. Defaults to :data:`True`.
    """

    __slots__ = (
        "_dimension",  # From base class
        "_nodes",  # From base class
        "_degree",  # From constructor
    )

    def __init__(self, nodes, degree, *, copy=True, verify=True):
        super(Curve, self).__init__(nodes, copy=copy)
        self._degree = degree
        self._verify_degree(verify)

    @classmethod
    def from_nodes(cls, nodes, copy=True):
        """Create a :class:`.Curve` from nodes.

        Computes the ``degree`` based on the shape of ``nodes``.

        Args:
            nodes (Sequence[Sequence[numbers.Number]]): The nodes in the curve.
                Must be convertible to a 2D NumPy array of floating point
                values, where the columns represent each node while the rows
                are the dimension of the ambient space.
            copy (bool): Flag indicating if the nodes should be copied before
                being stored. Defaults to :data:`True` since callers may
                freely mutate ``nodes`` after passing in.

        Returns:
            Curve: The constructed curve.
        """
        nodes_np = _base.sequence_to_array(nodes)
        _, num_nodes = nodes_np.shape
        degree = cls._get_degree(num_nodes)
        return cls(nodes_np, degree, copy=copy, verify=False)

    @staticmethod
    def _get_degree(num_nodes):
        """Get the degree of the current curve.

        Args:
            num_nodes (int): The number of nodes provided.

        Returns:
            int: The degree of the current curve.
        """
        return num_nodes - 1

    def _verify_degree(self, verify):
        """Verify that the number of nodes matches the degree.

        Args:
            verify (bool): Flag indicating if the degree should be verified
                against the number of nodes.

        Raises:
            ValueError: If ``verify`` is :data:`True` and the number of nodes
                does not match the degree.
        """
        if not verify:
            return

        _, num_nodes = self._nodes.shape
        expected_nodes = self._degree + 1
        if expected_nodes == num_nodes:
            return

        msg = (
            f"A degree {self._degree} curve should have "
            f"{expected_nodes} nodes, not {num_nodes}."
        )
        raise ValueError(msg)

    @property
    def length(self):
        r"""The length of the current curve.

        Computes the length via:

        .. math::

           \int_{B\left(\left[0, 1\right]\right)} 1 \, d\mathbf{x} =
           \int_0^1 \left\lVert B'(s) \right\rVert_2 \, ds

        Returns:
            float: The length of the current curve.
        """
        return _curve_helpers.compute_length(self._nodes)

    @property
    def __dict__(self):
        """dict: Dictionary of current curve's property namespace.

        This is just a stand-in property for the usual ``__dict__``. This
        class defines ``__slots__`` so by default would not provide a
        ``__dict__``.

        This also means that the current object can't be modified by the
        returned dictionary.
        """
        return {
            "_dimension": self._dimension,
            "_nodes": self._nodes,
            "_degree": self._degree,
        }

    def copy(self):
        """Make a copy of the current curve.

        Returns:
            .Curve: Copy of current curve.
        """
        return Curve(self._nodes, self._degree, copy=True, verify=False)

    def evaluate(self, s):
        r"""Evaluate :math:`B(s)` along the curve.

        This method acts as a (partial) inverse to :meth:`locate`.

        See :meth:`evaluate_multi` for more details.

        .. image:: ../../images/curve_evaluate.png
           :align: center

        .. doctest:: curve-eval
           :options: +NORMALIZE_WHITESPACE

           >>> nodes = np.asfortranarray([
           ...     [0.0, 0.625, 1.0],
           ...     [0.0, 0.5  , 0.5],
           ... ])
           >>> curve = bezier.Curve(nodes, degree=2)
           >>> curve.evaluate(0.75)
           array([[0.796875],
                  [0.46875 ]])

        .. testcleanup:: curve-eval

           import make_images
           make_images.curve_evaluate(curve)

        Args:
            s (float): Parameter along the curve.

        Returns:
            numpy.ndarray: The point on the curve (as a two dimensional
            NumPy array with a single column).
        """
        return _curve_helpers.evaluate_multi(
            self._nodes, np.asfortranarray([s])
        )

    def evaluate_multi(self, s_vals):
        r"""Evaluate :math:`B(s)` for multiple points along the curve.

        This is done via a modified Horner's method (vectorized for
        each ``s``-value).

        .. doctest:: curve-eval-multi
           :options: +NORMALIZE_WHITESPACE

           >>> nodes = np.asfortranarray([
           ...     [0.0, 1.0],
           ...     [0.0, 2.0],
           ...     [0.0, 3.0],
           ... ])
           >>> curve = bezier.Curve(nodes, degree=1)
           >>> curve
           <Curve (degree=1, dimension=3)>
           >>> s_vals = np.linspace(0.0, 1.0, 5)
           >>> curve.evaluate_multi(s_vals)
           array([[0.  , 0.25, 0.5 , 0.75, 1.  ],
                  [0.  , 0.5 , 1.  , 1.5 , 2.  ],
                  [0.  , 0.75, 1.5 , 2.25, 3.  ]])

        Args:
            s_vals (numpy.ndarray): Parameters along the curve (as a
                1D array).

        Returns:
            numpy.ndarray: The points on the curve. As a two dimensional
            NumPy array, with the columns corresponding to each ``s``
            value and the rows to the dimension.
        """
        return _curve_helpers.evaluate_multi(self._nodes, s_vals)

    def plot(self, num_pts, color=None, alpha=None, ax=None):
        """Plot the current curve.

        Args:
            num_pts (int): Number of points to plot.
            color (Optional[Tuple[float, float, float]]): Color as RGB profile.
            alpha (Optional[float]): The alpha channel for the color.
            ax (Optional[matplotlib.artist.Artist]): matplotlib axis object
                to add plot to.

        Returns:
            matplotlib.artist.Artist: The axis containing the plot. This
            may be a newly created axis.

        Raises:
            NotImplementedError: If the curve's dimension is not ``2``.
        """
        if self._dimension != 2:
            raise NotImplementedError(
                "2D is the only supported dimension",
                "Current dimension",
                self._dimension,
            )

        s_vals = np.linspace(0.0, 1.0, num_pts)
        points = self.evaluate_multi(s_vals)
        if ax is None:
            ax = _plot_helpers.new_axis()
        ax.plot(points[0, :], points[1, :], color=color, alpha=alpha)
        return ax

    def subdivide(self):
        r"""Split the curve :math:`B(s)` into a left and right half.

        Takes the interval :math:`\left[0, 1\right]` and splits the curve into
        :math:`B_1 = B\left(\left[0, \frac{1}{2}\right]\right)` and
        :math:`B_2 = B\left(\left[\frac{1}{2}, 1\right]\right)`. In
        order to do this, also reparameterizes the curve, hence the resulting
        left and right halves have new nodes.

        .. image:: ../../images/curve_subdivide.png
           :align: center

        .. doctest:: curve-subdivide
           :options: +NORMALIZE_WHITESPACE

           >>> nodes = np.asfortranarray([
           ...     [0.0, 1.25, 2.0],
           ...     [0.0, 3.0 , 1.0],
           ... ])
           >>> curve = bezier.Curve(nodes, degree=2)
           >>> left, right = curve.subdivide()
           >>> left.nodes
           array([[0.   , 0.625, 1.125],
                  [0.   , 1.5  , 1.75 ]])
           >>> right.nodes
           array([[1.125, 1.625, 2.   ],
                  [1.75 , 2.   , 1.   ]])

        .. testcleanup:: curve-subdivide

           import make_images
           make_images.curve_subdivide(curve, left, right)

        Returns:
            Tuple[Curve, Curve]: The left and right sub-curves.
        """
        left_nodes, right_nodes = _curve_helpers.subdivide_nodes(self._nodes)
        left = Curve(left_nodes, self._degree, copy=False, verify=False)
        right = Curve(right_nodes, self._degree, copy=False, verify=False)
        return left, right

    def intersect(
        self, other, strategy=IntersectionStrategy.GEOMETRIC, _verify=True
    ):
        """Find the points of intersection with another curve.

        See :doc:`../../algorithms/curve-curve-intersection` for more details.

        .. image:: ../../images/curve_intersect.png
           :align: center

        .. doctest:: curve-intersect
           :options: +NORMALIZE_WHITESPACE

           >>> nodes1 = np.asfortranarray([
           ...     [0.0, 0.375, 0.75 ],
           ...     [0.0, 0.75 , 0.375],
           ... ])
           >>> curve1 = bezier.Curve(nodes1, degree=2)
           >>> nodes2 = np.asfortranarray([
           ...     [0.5, 0.5 ],
           ...     [0.0, 0.75],
           ... ])
           >>> curve2 = bezier.Curve(nodes2, degree=1)
           >>> intersections = curve1.intersect(curve2)
           >>> 3.0 * intersections
           array([[2.],
                  [2.]])
           >>> s_vals = intersections[0, :]
           >>> curve1.evaluate_multi(s_vals)
           array([[0.5],
                  [0.5]])

        .. testcleanup:: curve-intersect

           import make_images
           make_images.curve_intersect(curve1, curve2, s_vals)

        Args:
            other (Curve): Other curve to intersect with.
            strategy (Optional[~bezier.curve.IntersectionStrategy]): The
                intersection algorithm to use. Defaults to geometric.
            _verify (Optional[bool]): Indicates if extra caution should be
                used to verify assumptions about the input and current
                curve. Can be disabled to speed up execution time.
                Defaults to :data:`True`.

        Returns:
            numpy.ndarray: ``2 x N`` array of ``s``- and ``t``-parameters where
            intersections occur (possibly empty).

        Raises:
            TypeError: If ``other`` is not a curve (and ``_verify=True``).
            NotImplementedError: If at least one of the curves
                isn't two-dimensional (and ``_verify=True``).
            ValueError: If ``strategy`` is not a valid
                :class:`.IntersectionStrategy`.
        """
        if _verify:
            if not isinstance(other, Curve):
                raise TypeError(
                    "Can only intersect with another curve", "Received", other
                )

            if self._dimension != 2 or other._dimension != 2:
                raise NotImplementedError(
                    "Intersection only implemented in 2D"
                )

        if strategy == IntersectionStrategy.GEOMETRIC:
            all_intersections = _geometric_intersection.all_intersections
        elif strategy == IntersectionStrategy.ALGEBRAIC:
            all_intersections = _algebraic_intersection.all_intersections
        else:
            raise ValueError("Unexpected strategy.", strategy)

        st_vals, _ = all_intersections(self._nodes, other._nodes)
        return st_vals

    def elevate(self):
        r"""Return a degree-elevated version of the current curve.

        Does this by converting the current nodes :math:`v_0, \ldots, v_n`
        to new nodes :math:`w_0, \ldots, w_{n + 1}` where

        .. math::

           \begin{align*}
           w_0 &= v_0 \\
           w_j &= \frac{j}{n + 1} v_{j - 1} + \frac{n + 1 - j}{n + 1} v_j \\
           w_{n + 1} &= v_n
           \end{align*}

        .. image:: ../../images/curve_elevate.png
           :align: center

        .. testsetup:: curve-elevate

           import numpy as np
           import bezier

        .. doctest:: curve-elevate
           :options: +NORMALIZE_WHITESPACE

           >>> nodes = np.asfortranarray([
           ...     [0.0, 1.5, 3.0],
           ...     [0.0, 1.5, 0.0],
           ... ])
           >>> curve = bezier.Curve(nodes, degree=2)
           >>> elevated = curve.elevate()
           >>> elevated
           <Curve (degree=3, dimension=2)>
           >>> elevated.nodes
           array([[0., 1., 2., 3.],
                  [0., 1., 1., 0.]])

        .. testcleanup:: curve-elevate

           import make_images
           make_images.curve_elevate(curve, elevated)

        Returns:
            Curve: The degree-elevated curve.
        """
        new_nodes = _curve_helpers.elevate_nodes(self._nodes)
        return Curve(new_nodes, self._degree + 1, copy=False, verify=False)

    def reduce_(self):
        r"""Return a degree-reduced version of the current curve.

        .. _pseudo-inverse:
            https://en.wikipedia.org/wiki/Moore%E2%80%93Penrose_pseudoinverse

        Does this by converting the current nodes :math:`v_0, \ldots, v_n`
        to new nodes :math:`w_0, \ldots, w_{n - 1}` that correspond to
        reversing the :meth:`elevate` process.

        This uses the `pseudo-inverse`_ of the elevation matrix. For example
        when elevating from degree 2 to 3, the matrix :math:`E_2` is given by

        .. math::

           \mathbf{v} = \left[\begin{array}{c c c} v_0 & v_1 & v_2
               \end{array}\right] \longmapsto \left[\begin{array}{c c c c}
               v_0 & \frac{v_0 + 2 v_1}{3} & \frac{2 v_1 + v_2}{3} & v_2
               \end{array}\right] = \frac{1}{3} \mathbf{v}
               \left[\begin{array}{c c c c} 3 & 1 & 0 & 0 \\
               0 & 2 & 2 & 0 \\ 0 & 0 & 1 & 3 \end{array}\right]

        and the (right) pseudo-inverse is given by

        .. math::

           R_2 = E_2^T \left(E_2 E_2^T\right)^{-1} = \frac{1}{20}
               \left[\begin{array}{c c c} 19 & -5 & 1 \\
               3 & 15 & -3 \\ -3 & 15 & 3 \\ 1 & -5 & 19
               \end{array}\right].

        .. warning::

           Though degree-elevation preserves the start and end nodes, degree
           reduction has no such guarantee. Rather, the nodes produced are
           "best" in the least squares sense (when solving the normal
           equations).

        .. image:: ../../images/curve_reduce.png
           :align: center

        .. testsetup:: curve-reduce, curve-reduce-approx

           import numpy as np
           import bezier

        .. doctest:: curve-reduce
           :options: +NORMALIZE_WHITESPACE

           >>> nodes = np.asfortranarray([
           ...     [-3.0, 0.0, 1.0, 0.0],
           ...     [ 3.0, 2.0, 3.0, 6.0],
           ... ])
           >>> curve = bezier.Curve(nodes, degree=3)
           >>> reduced = curve.reduce_()
           >>> reduced
           <Curve (degree=2, dimension=2)>
           >>> reduced.nodes
           array([[-3. ,  1.5,  0. ],
                  [ 3. ,  1.5,  6. ]])

        .. testcleanup:: curve-reduce

           import make_images
           make_images.curve_reduce(curve, reduced)

        In the case that the current curve **is not** degree-elevated.

        .. image:: ../../images/curve_reduce_approx.png
           :align: center

        .. doctest:: curve-reduce-approx
           :options: +NORMALIZE_WHITESPACE

           >>> nodes = np.asfortranarray([
           ...     [0.0, 1.25, 3.75, 5.0],
           ...     [2.5, 5.0 , 7.5 , 2.5],
           ... ])
           >>> curve = bezier.Curve(nodes, degree=3)
           >>> reduced = curve.reduce_()
           >>> reduced
           <Curve (degree=2, dimension=2)>
           >>> reduced.nodes
           array([[-0.125,  2.5  ,  5.125],
                  [ 2.125,  8.125,  2.875]])

        .. testcleanup:: curve-reduce-approx

           import make_images
           make_images.curve_reduce_approx(curve, reduced)

        Returns:
            Curve: The degree-reduced curve.
        """
        new_nodes = _curve_helpers.reduce_pseudo_inverse(self._nodes)
        return Curve(new_nodes, self._degree - 1, copy=False, verify=False)

    def specialize(self, start, end):
        """Specialize the curve to a given sub-interval.

        .. image:: ../../images/curve_specialize.png
           :align: center

        .. doctest:: curve-specialize

           >>> nodes = np.asfortranarray([
           ...     [0.0, 0.5, 1.0],
           ...     [0.0, 1.0, 0.0],
           ... ])
           >>> curve = bezier.Curve(nodes, degree=2)
           >>> new_curve = curve.specialize(-0.25, 0.75)
           >>> new_curve.nodes
           array([[-0.25 ,  0.25 ,  0.75 ],
                  [-0.625,  0.875,  0.375]])

        .. testcleanup:: curve-specialize

           import make_images
           make_images.curve_specialize(curve, new_curve)

        This is generalized version of :meth:`subdivide`, and can even
        match the output of that method:

        .. testsetup:: curve-specialize2

           import numpy as np
           import bezier

           nodes = np.asfortranarray([
               [0.0, 0.5, 1.0],
               [0.0, 1.0, 0.0],
           ])
           curve = bezier.Curve(nodes, degree=2)

        .. doctest:: curve-specialize2

           >>> left, right = curve.subdivide()
           >>> also_left = curve.specialize(0.0, 0.5)
           >>> np.all(also_left.nodes == left.nodes)
           True
           >>> also_right = curve.specialize(0.5, 1.0)
           >>> np.all(also_right.nodes == right.nodes)
           True

        Args:
            start (float): The start point of the interval we
                are specializing to.
            end (float): The end point of the interval we
                are specializing to.

        Returns:
            Curve: The newly-specialized curve.
        """
        new_nodes = _curve_helpers.specialize_curve(self._nodes, start, end)
        return Curve(new_nodes, self._degree, copy=False, verify=False)

    def locate(self, point):
        r"""Find a point on the current curve.

        Solves for :math:`s` in :math:`B(s) = p`.

        This method acts as a (partial) inverse to :meth:`evaluate`.

        .. note::

           A unique solution is only guaranteed if the current curve has no
           self-intersections. This code assumes, but doesn't check, that
           this is true.

        .. image:: ../../images/curve_locate.png
           :align: center

        .. doctest:: curve-locate

           >>> nodes = np.asfortranarray([
           ...     [0.0, -1.0, 1.0, -0.75 ],
           ...     [2.0,  0.0, 1.0,  1.625],
           ... ])
           >>> curve = bezier.Curve(nodes, degree=3)
           >>> point1 = np.asfortranarray([
           ...     [-0.09375 ],
           ...     [ 0.828125],
           ... ])
           >>> curve.locate(point1)
           0.5
           >>> point2 = np.asfortranarray([
           ...     [0.0],
           ...     [1.5],
           ... ])
           >>> curve.locate(point2) is None
           True
           >>> point3 = np.asfortranarray([
           ...     [-0.25 ],
           ...     [ 1.375],
           ... ])
           >>> curve.locate(point3) is None
           Traceback (most recent call last):
             ...
           ValueError: Parameters not close enough to one another

        .. testcleanup:: curve-locate

           import make_images
           make_images.curve_locate(curve, point1, point2, point3)

        Args:
            point (numpy.ndarray): A (``D x 1``) point on the curve,
                where :math:`D` is the dimension of the curve.

        Returns:
            Optional[float]: The parameter value (:math:`s`) corresponding
            to ``point`` or :data:`None` if the point is not on
            the ``curve``.

        Raises:
            ValueError: If the dimension of the ``point`` doesn't match the
                dimension of the current curve.
        """
        if point.shape != (self._dimension, 1):
            point_dimensions = " x ".join(
                str(dimension) for dimension in point.shape
            )
            msg = _LOCATE_ERROR_TEMPLATE.format(
                self._dimension, self._dimension, point, point_dimensions
            )
            raise ValueError(msg)

        return _curve_helpers.locate_point(self._nodes, point)

    # Return type doc appears missing to Pylint because of the use of the
    # :class:`sympy.Matrix ...` aliases.
    # pylint: disable=missing-return-type-doc
    def to_symbolic(self):
        """Convert to a SymPy matrix representing :math:`B(s)`.

        .. note::

           This method requires SymPy.

        .. doctest:: curve-to-symbolic

           >>> nodes = np.asfortranarray([
           ...     [0.0, -1.0, 1.0, -0.75 ],
           ...     [2.0,  0.0, 1.0,  1.625],
           ... ])
           >>> curve = bezier.Curve(nodes, degree=3)
           >>> curve.to_symbolic()
           Matrix([
           [               -3*s*(3*s - 2)**2/4],
           [-(27*s**3 - 72*s**2 + 48*s - 16)/8]])

        Returns:
            :class:`sympy.Matrix <sympy.matrices.dense.MutableDenseMatrix>`:
            The curve :math:`B(s)`.
        """
        _, b_polynomial = _symbolic.curve_as_polynomial(
            self._nodes, self._degree
        )
        return b_polynomial

    def implicitize(self):
        r"""Implicitize the curve .

        .. note::

           This method requires SymPy.

        .. doctest:: curve-implicitize

           >>> nodes = np.asfortranarray([
           ...     [0.0, 1.0, 1.0],
           ...     [2.0, 0.0, 1.0],
           ... ])
           >>> curve = bezier.Curve(nodes, degree=2)
           >>> curve.implicitize()
           9*x**2 + 6*x*y - 20*x + y**2 - 8*y + 12

        Returns:
            :class:`sympy.Expr <sympy.core.expr.Expr>`: The function that
            defines the curve in :math:`\mathbf{R}^2` via :math:`f(x, y) = 0`.

        Raises:
            ValueError: If the curve's dimension is not ``2``.
        """
        if self._dimension != 2:
            raise ValueError(
                "Only a planar (2D) curve can be implicitized",
                "Current dimension",
                self._dimension,
            )

        return _symbolic.implicitize_curve(self._nodes, self._degree)

    # pylint: enable=missing-return-type-doc



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Cython wrapper for ``status.f90``."""


cdef extern from "bezier/status.h":
    cpdef enum Status:
        SUCCESS = 0
        BAD_MULTIPLICITY = 1
        NO_CONVERGE = 2
        INSUFFICIENT_SPACE = 3
        SAME_CURVATURE = 4
        BAD_INTERIOR = 5
        EDGE_END = 6
        SINGULAR = 7
        UNKNOWN = 999



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""Proof-of-concept for B |eacute| zier clipping.

.. _algorithm: https://dx.doi.org/10.1016/0010-4485(90)90039-F

The B |eacute| zier clipping `algorithm`_ is used to intersect
two planar B |eacute| zier curves. It proceeds by using "fat lines"
to recursively prune the region of accepted parameter ranges until
the ranges converge to points. (A "fat line" is a rectangular region of a
bounded distance from the line connecting the start and end points of a
B |eacute| zier curve.)

It has quadratic convergence. It can be used to find tangent intersections,
which is the primary usage within ``bezier``.

.. |eacute| unicode:: U+000E9 .. LATIN SMALL LETTER E WITH ACUTE
   :trim:
"""

import numpy as np

from bezier import _helpers
from bezier import _py_geometric_intersection


NO_PARALLEL = "Parallel lines not supported during clipping."
DEFAULT_S_MIN = 1.0
DEFAULT_S_MAX = 0.0


def compute_implicit_line(nodes):
    """Compute the implicit form of the line connecting curve endpoints.

    .. note::

       This assumes, but does not check, that the first and last nodes
       in ``nodes`` are different.

    Computes :math:`a, b` and :math:`c` in the normalized implicit equation
    for the line

    .. math::

       ax + by + c = 0

    where :math:`a^2 + b^2 = 1` (only unique up to sign).

    Args:
        nodes (numpy.ndarray): ``2 x N`` array of nodes in a curve.
            The line will be (directed) from the first to last
            node in ``nodes``.

    Returns:
        Tuple[float, float, float]: The triple of

        * The :math:`x` coefficient :math:`a`
        * The :math:`y` coefficient :math:`b`
        * The constant :math:`c`
    """
    delta = nodes[:, -1] - nodes[:, 0]
    length = np.linalg.norm(delta, ord=2)
    # Normalize and rotate 90 degrees to the "left".
    coeff_a = -delta[1] / length
    coeff_b = delta[0] / length
    # c = - ax - by = (delta[1] x - delta[0] y) / L
    # NOTE: We divide by ``length`` at the end to "put off" rounding.
    coeff_c = (delta[1] * nodes[0, 0] - delta[0] * nodes[1, 0]) / length
    return coeff_a, coeff_b, coeff_c


def compute_fat_line(nodes):
    """Compute the "fat line" around a B |eacute| zier curve.

    Both computes the implicit (normalized) form

    .. math::

       ax + by + c = 0

    for the line connecting the first and last node in ``nodes``.
    Also computes the maximum and minimum distances to that line
    from each control point.

    Args:
        nodes (numpy.ndarray): ``2 x N`` array of nodes in a curve.

    Returns:
        Tuple[float, float, float, float, float]: The 5-tuple of

        * The :math:`x` coefficient :math:`a`
        * The :math:`y` coefficient :math:`b`
        * The constant :math:`c`
        * The "minimum" distance to the fat line among the control points.
        * The "maximum" distance to the fat line among the control points.
    """
    coeff_a, coeff_b, coeff_c = compute_implicit_line(nodes)
    # NOTE: This assumes, but does not check, that there are two rows.
    _, num_nodes = nodes.shape
    d_min = 0.0
    d_max = 0.0
    for index in range(1, num_nodes - 1):  # Only interior nodes.
        curr_dist = (
            coeff_a * nodes[0, index] + coeff_b * nodes[1, index] + coeff_c
        )
        if curr_dist < d_min:
            d_min = curr_dist
        elif curr_dist > d_max:
            d_max = curr_dist
    return coeff_a, coeff_b, coeff_c, d_min, d_max


def _update_parameters(s_min, s_max, start0, end0, start1, end1):
    """Update clipped parameter range.

    .. note::

       This is a helper for :func:`clip_range`.

    Does so by intersecting one of the two fat lines with an edge
    of the convex hull of the distance polynomial of the curve being
    clipped.

    If both of ``s_min`` and ``s_max`` are "unset", then any :math:`s`
    value that is valid for ``s_min`` would also be valid for ``s_max``.
    Rather than adding a special case to handle this scenario, **only**
    ``s_min`` will be updated.

    In cases where a given parameter :math:`s` would be a valid update
    for both ``s_min`` and ``s_max``
    This function **only** updates ``s_min``

    Args:
        s_min (float): Current start of clipped interval. If "unset", this
            value will be ``DEFAULT_S_MIN``.
        s_max (float): Current end of clipped interval. If "unset", this
            value will be ``DEFAULT_S_MAX``.
        start0 (numpy.ndarray): A 1D NumPy ``2``-array that is the start
            vector of one of the two fat lines.
        end0 (numpy.ndarray): A 1D NumPy ``2``-array that is the end
            vector of one of the two fat lines.
        start1 (numpy.ndarray): A 1D NumPy ``2``-array that is the start
            vector of an edge of the convex hull of the distance
            polynomial :math:`d(t)` as an explicit B |eacute| zier curve.
        end1 (numpy.ndarray): A 1D NumPy ``2``-array that is the end
            vector of an edge of the convex hull of the distance
            polynomial :math:`d(t)` as an explicit B |eacute| zier curve.

    Returns:
        Tuple[float, float]: The (possibly updated) start and end
        of the clipped parameter range.

    Raises:
        NotImplementedError: If the two line segments are parallel. (This
            case will be supported at some point, just not now.)
    """
    s, t, success = _py_geometric_intersection.segment_intersection(
        start0, end0, start1, end1
    )

    if not success:
        raise NotImplementedError(NO_PARALLEL)

    if _helpers.in_interval(t, 0.0, 1.0):
        if _helpers.in_interval(s, 0.0, s_min):
            return s, s_max

        elif _helpers.in_interval(s, s_max, 1.0):
            return s_min, s

    return s_min, s_max


def _check_parameter_range(s_min, s_max):
    r"""Performs a final check on a clipped parameter range.

    .. note::

       This is a helper for :func:`clip_range`.

    If both values are unchanged from the "unset" default, this returns
    the whole interval :math:`\left[0.0, 1.0\right]`.

    If only one of the values is set to some parameter :math:`s`, this
    returns the "degenerate" interval :math:`\left[s, s\right]`. (We rely
    on the fact that ``s_min`` must be the only set value, based on how
    :func:`_update_parameters` works.)

    Otherwise, this simply returns ``[s_min, s_max]``.

    Args:
        s_min (float): Current start of clipped interval. If "unset", this
            value will be ``DEFAULT_S_MIN``.
        s_max (float): Current end of clipped interval. If "unset", this
            value will be ``DEFAULT_S_MAX``.

    Returns:
        Tuple[float, float]: The (possibly updated) start and end
        of the clipped parameter range.
    """
    if s_min == DEFAULT_S_MIN:
        # Based on the way ``_update_parameters`` works, we know
        # both parameters must be unset if ``s_min``.
        return 0.0, 1.0

    if s_max == DEFAULT_S_MAX:
        return s_min, s_min

    return s_min, s_max


def clip_range(nodes1, nodes2):
    r"""Reduce the parameter range where two curves can intersect.

    Does so by using the "fat line" for ``nodes1`` and computing the
    distance polynomial against ``nodes2``.

    .. note::

       This assumes, but does not check that the curves being considered
       will only have one intersection in the parameter ranges
       :math:`s \in \left[0, 1\right]`, :math:`t \in \left[0, 1\right]`.
       This assumption is based on the fact that B |eacute| zier clipping
       is meant to be used to find tangent intersections for already
       subdivided (i.e. sufficiently zoomed in) curve segments.

    Args:
        nodes1 (numpy.ndarray): ``2 x N1`` array of nodes in a curve which
            will define the clipping region.
        nodes2 (numpy.ndarray): ``2 x N2`` array of nodes in a curve which
            will be clipped.

    Returns:
        Tuple[float, float]: The pair of

        * The start parameter of the clipped range.
        * The end parameter of the clipped range.
    """
    # NOTE: There is no corresponding "enable", but the disable only applies
    #       in this lexical scope.
    # pylint: disable=too-many-locals
    coeff_a, coeff_b, coeff_c, d_min, d_max = compute_fat_line(nodes1)
    # NOTE: This assumes, but does not check, that there are two rows.
    _, num_nodes2 = nodes2.shape
    polynomial = np.empty((2, num_nodes2), order="F")
    denominator = float(num_nodes2 - 1)
    for index in range(num_nodes2):
        polynomial[0, index] = index / denominator
        polynomial[1, index] = (
            coeff_a * nodes2[0, index] + coeff_b * nodes2[1, index] + coeff_c
        )
    # Define segments for the top and the bottom of the region
    # bounded by the fat line.
    start_bottom = np.asfortranarray([0.0, d_min])
    end_bottom = np.asfortranarray([1.0, d_min])
    start_top = np.asfortranarray([0.0, d_max])
    end_top = np.asfortranarray([1.0, d_max])
    s_min = DEFAULT_S_MIN
    s_max = DEFAULT_S_MAX
    # NOTE: We avoid computing the convex hull and just compute where
    #       all segments connecting two control points intersect the
    #       fat lines.
    for start_index in range(num_nodes2 - 1):
        for end_index in range(start_index + 1, num_nodes2):
            s_min, s_max = _update_parameters(
                s_min,
                s_max,
                start_bottom,
                end_bottom,
                polynomial[:, start_index],
                polynomial[:, end_index],
            )
            s_min, s_max = _update_parameters(
                s_min,
                s_max,
                start_top,
                end_top,
                polynomial[:, start_index],
                polynomial[:, end_index],
            )
    return _check_parameter_range(s_min, s_max)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Plotting utilities."""

import numpy as np

from bezier import _helpers


def new_axis():
    """Get a new matplotlib axis.

    Returns:
        matplotlib.artist.Artist: A newly created axis.
    """
    # NOTE: We import the plotting library at runtime to
    #       avoid the cost for users that only want to compute.
    #       The ``matplotlib`` import is a tad expensive.
    import matplotlib.pyplot as plt  # pylint: disable=import-outside-toplevel

    figure = plt.figure()
    return figure.gca()


def add_plot_boundary(ax, padding=0.125):
    """Add a buffer of empty space around a plot boundary.

    .. note::

       This only uses ``line`` data from the axis. It **could**
       use ``patch`` data, but doesn't at this time.

    Args:
        ax (matplotlib.artist.Artist): A matplotlib axis.
        padding (Optional[float]): Amount (as a fraction of width and height)
            of padding to add around data. Defaults to ``0.125``.
    """
    nodes = np.asfortranarray(
        np.vstack([line.get_xydata() for line in ax.lines]).T
    )
    left, right, bottom, top = _helpers.bbox(nodes)
    center_x = 0.5 * (right + left)
    delta_x = right - left
    center_y = 0.5 * (top + bottom)
    delta_y = top - bottom
    multiplier = (1.0 + padding) * 0.5
    ax.set_xlim(
        center_x - multiplier * delta_x, center_x + multiplier * delta_x
    )
    ax.set_ylim(
        center_y - multiplier * delta_y, center_y + multiplier * delta_y
    )


def add_patch(ax, color, pts_per_edge, *edges):
    """Add a polygonal surface patch to a plot.

    Args:
        ax (matplotlib.artist.Artist): A matplotlib axis.
        color (Tuple[float, float, float]): Color as RGB profile.
        pts_per_edge (int): Number of points to use in polygonal
            approximation of edge.
        edges (Tuple[~bezier.curve.Curve, ...]): Curved edges defining
            a boundary.
    """
    # pylint: disable=import-outside-toplevel
    from matplotlib import patches
    from matplotlib import path as _path_mod

    # pylint: enable=import-outside-toplevel

    s_vals = np.linspace(0.0, 1.0, pts_per_edge)
    # Evaluate points on each edge.
    all_points = []
    for edge in edges:
        points = edge.evaluate_multi(s_vals)
        # We assume the edges overlap and leave out the first point
        # in each.
        all_points.append(points[:, 1:])
    # Add first point as last point (polygon is closed).
    first_edge = all_points[0]
    all_points.append(first_edge[:, [0]])
    # Add boundary first.
    polygon = np.asfortranarray(np.hstack(all_points))
    (line,) = ax.plot(polygon[0, :], polygon[1, :], color=color)
    # Reset ``color`` in case it was ``None`` and set from color wheel.
    color = line.get_color()
    # ``polygon`` is stored Fortran-contiguous with ``x-y`` points in each
    # column but ``Path()`` wants ``x-y`` points in each row.
    path = _path_mod.Path(polygon.T)
    patch = patches.PathPatch(path, facecolor=color, alpha=0.625)
    ax.add_patch(patch)
